/* OpenCA OCSPD - src/openssl_threads_help.c
 * ====================================================================
 * Threads help function from OpenSSL - slightly modified by Madwolf
 */

#include "general.h"

#ifndef CRYPTO_LOCK
#define CRYPTO_LOCK     0x01
#define CRYPTO_UNLOCK   0x02
#define CRYPTO_READ     0x04
#define CRYPTO_WRITE    0x08
#endif

struct CRYPTO_dynlock_value {
	pthread_rwlock_t lock;        
};

static pthread_mutex_t *lock_cs;
static long *lock_count;

/* Local Function Declaration */
void _dyn_lock_callback(int mode, struct CRYPTO_dynlock_value *l, 
						const char *file, int line);
struct CRYPTO_dynlock_value *_dyn_create_callback(const char *file, int line);
void _dyn_destroy_callback(struct CRYPTO_dynlock_value *l, 
						const char *file, int line);
void thread_cleanup(void);
void pthreads_locking_callback(int mode, int type, char *file, int line);
unsigned long pthreads_thread_id(void);

/* Functions */

void _dyn_lock_callback(int mode, struct CRYPTO_dynlock_value *l, 
						const char *file, int line) {
	     if(mode==(CRYPTO_LOCK|CRYPTO_READ))
		pthread_rwlock_rdlock(&l->lock);
	else if(mode==(CRYPTO_LOCK|CRYPTO_WRITE)) 
		pthread_rwlock_wrlock(&l->lock);
	else if(mode==(CRYPTO_UNLOCK|CRYPTO_READ))
		pthread_rwlock_unlock(&l->lock);
	else if(mode==(CRYPTO_UNLOCK|CRYPTO_WRITE))
		pthread_rwlock_unlock(&l->lock);

	return;
}
 
struct CRYPTO_dynlock_value *_dyn_create_callback(const char *file, int line) {
	struct CRYPTO_dynlock_value *l = NULL;

	l = (struct CRYPTO_dynlock_value *) 
				malloc(sizeof(struct CRYPTO_dynlock_value));
	pthread_rwlock_init(&l->lock, NULL);

	return l;
}

void _dyn_destroy_callback(struct CRYPTO_dynlock_value *l, 
						const char *file, int line) {
	pthread_rwlock_destroy(&l->lock);
	free(l);
}
                                  
void OpenSSL_thread_setup(void) {
	int i;

	lock_cs=OPENSSL_malloc(CRYPTO_num_locks() * sizeof(pthread_mutex_t));
	lock_count=OPENSSL_malloc(CRYPTO_num_locks() * sizeof(long));
	for (i=0; i<CRYPTO_num_locks(); i++)
		{
		lock_count[i]=0;
		pthread_mutex_init(&(lock_cs[i]),NULL);
		}

	CRYPTO_set_id_callback((unsigned long (*)())pthreads_thread_id);
	CRYPTO_set_locking_callback((void (*)())pthreads_locking_callback);

	/* Initializing the OpenSSL dynamic callbacks as well,
           needed by the nCipher driver */
	CRYPTO_set_dynlock_create_callback(_dyn_create_callback);
	CRYPTO_set_dynlock_lock_callback(_dyn_lock_callback);
	CRYPTO_set_dynlock_destroy_callback(_dyn_destroy_callback);

	return;
}

void thread_cleanup(void) {
	int i;

	CRYPTO_set_locking_callback(NULL);

	for (i=0; i<CRYPTO_num_locks(); i++) {
		pthread_mutex_destroy(&(lock_cs[i]));
	}

	OPENSSL_free(lock_cs);
	OPENSSL_free(lock_count);

	return;
}

void pthreads_locking_callback(int mode, int type, char *file, int line) {

	if (mode & CRYPTO_LOCK) {
		pthread_mutex_lock(&(lock_cs[type]));
		lock_count[type]++;
	} else {
		pthread_mutex_unlock(&(lock_cs[type]));
	}

	return;
}

unsigned long pthreads_thread_id(void) {
	unsigned long ret;

	ret=(unsigned long)pthread_self();
	return(ret);
}

